package edu.cosy.netgen.ann;

import java.util.*;

public class FeedForwardNet extends NeuralNet {
    
    // description of the net
    private final String DESCRIPTION = "FeedForwardNet";
    
    // type of the net for random created nets
    private int nettype;

    // 1 (bias) + input + hidden +output neurons
    private int matrixWidth = 0;
    
    public FeedForwardNet() {
        this.in = 0;
        this.hidden = 0;
        this.out = 0;
        this.conMatrix = new ConnectionMatrix();
        this.weightMatrix = new WeightMatrix();
        this.actFuncNeuron = new ActFunction[]{ new ActFunctionSigmoid() };
    }
    
    /**
      * A new net instance, user specified net
      * @param in number of input neurons
      * @param hidden number of hidden neurons
      * @param out number of output neurons
      * @param markersModus use markers?
      */
    public FeedForwardNet( int in, int hidden, int out ) {
        this.in = in;
        this.out = out;
        this.hidden = hidden;
        this.matrixWidth = 1 + in + hidden + out;
    }

    /**
      * There are some ways to initialize a net, random nets are made
      * @param actFunc act. func. pool for hidden neurons
      * @param actFuncIn act. func. pool for input neurons
      * @param actFuncOut act. func. pool for output neurons
      * @param nettype which random net to be created
      * @return false, if net cannot initialized
      */
    public boolean initNet( ActFunction[] actFunc, ActFunction[] actFuncIn, ActFunction[] actFuncOut, int nettype ) {
        boolean retVal = false;
        this.nettype = nettype;

        conMatrix = new ConnectionMatrix( in, hidden, out, nettype );
        weightMatrix = new WeightMatrix( in, hidden, out );
        actFuncNeuron = new ActFunction[ matrixWidth ];

        //makes the random net
        return createNet( actFunc, actFuncIn, actFuncOut, nettype );
    }

    /**
      * There are some ways to initialize a net, act. func pools are generated
      * @param conMatrix connection matrix
      * @param weightMatrix weight matrix
      * @param markers marker for input and hidden neurons
      * @param actFuncNeuron for each neuron a own act. func.
      * @return false, if net cannot initialized
      */
    public boolean initNet( ConnectionMatrix conMatrix, WeightMatrix weightMatrix, boolean[] markers, ActFunction[] actFuncNeuron ) {
        boolean retVal = false;
        if ( conMatrix != null && weightMatrix != null && markers != null && actFuncNeuron != null ) {
            this.conMatrix = conMatrix;
            this.weightMatrix = weightMatrix;            
            conMatrix.setMarkers( markers );
            this.actFuncNeuron = actFuncNeuron;

            retVal = true;
        }
        return retVal;
    }

    /**
     * Get the description of the network
     * @return description
     */
    public String getDescription() {
        return DESCRIPTION;
    }

    // create a network
    private boolean createNet( ActFunction[] actFunc, ActFunction[] actFuncIn, ActFunction[] actFuncOut, int nettype ) {
        boolean retVal = false;
        if ( in > 0 && out > 0 && hidden > 0 ) {
            
            weightMatrix.createRandomWeights( conMatrix );
                        
            //kill a neuron with a propability of 50%
            if ( markersModus ) {
                boolean[] markers = new boolean[ conMatrix.size() + 1 ];
                for ( int i = 1; i < markers.length; i++ ) {
                    if ( Math.random() < 0.5 )
                        markers[ i ] = false;
                }
                conMatrix.setMarkers( markers );
            }

            //set input act. func.
            for ( int i = 1; i < 1 + in; i++ ) {
                actFuncNeuron[ i ] = actFuncIn[ ( int ) ( Math.random() * actFunc.length ) ];
            }
            
            //set hidden act. func.
            for ( int i = 1 + in; i < 1 + in + hidden; i++ ) {
                actFuncNeuron[ i ] = actFunc[ ( int ) ( Math.random() * actFunc.length ) ];
            }
                                
            //set output act. func.
            for ( int i = 1 + in + hidden; i < matrixWidth; i++ ) {
                actFuncNeuron[ i ] = actFuncOut[ ( int ) ( Math.random() * actFunc.length ) ];
            }
        }
        return retVal;
    }
    
    public String toString() {
        StringBuffer retVal = new StringBuffer( "Neural Network:\n\n" );
        retVal.append( "Description:\t" ).append( getDescription() );
        retVal.append( "\nInput Neurons:\t" ).append( getIn() );
        retVal.append( "\nHidden Neurons:\t" ).append( getHidden() );
        retVal.append( "\nOutput Neurons:\t" ).append( getOut() );
        retVal.append( "\nMarkers Modus:\t" ).append( markersModus ).append( " (" ).append( conMatrix.getMarkers() != null ? conMatrix.getMarkers().length : 0 ).append( ")" );
        retVal.append( "\n" ).append( getConMatrix() );
        retVal.append( "\n" ).append( getWeightMatrix() );
        return retVal.toString();
    }
}
