package edu.cosy.netgen.ann;


public class WeightMatrix implements java.io.Serializable {
    
    /**
     * Connection weights.
     */
    protected double[][] weights;
    
    /**
     * Number of input, hidden and output neurons.
     */
    protected int input, hidden, output;
    
    /**
     * Initialize a weight matrix with none neuron.
     */
    public WeightMatrix() {
        this( 0, 0, 0 );
    }
    
    /**
     * Copy constructor for the class.
     * <P>
     * @param wm weight matrix to copy
     */
    public WeightMatrix( WeightMatrix wm ) {
        super();
        weights = new double[ wm.weights.length ][ wm.weights.length ];
        System.arraycopy( wm.weights, 0, weights, 0, weights.length );
        this.input = wm.input;
        this.hidden = wm.hidden;
        this.output = wm.output;
    }
    
    /**
     * Create a new weight matrix. All weights are initialized with the lowest double value.
     * <P>
     * @param in number of input neurons
     * @param hidden number of hidden neurons
     * @param out number of output neurons
     */
    public WeightMatrix( int in, int hidden, int out ) {
        super();
        this.input = in;
        this.hidden = hidden;
        this.output = out;
        weights = new double[ 1 + input + hidden + output ][ 1 + input + hidden + output ];
            for ( int i = 0; i < weights.length; i++ )
                for ( int j = 0; j < weights.length; j++ )
                    weights[ i ][ j ] = Double.longBitsToDouble( 0x1L );
    }
    
    /**
     * Set a new weigtht for a connection.
     * <P>
     * @param s source neuron
     * @param d destination neuron
     * @param w new weight for the connection
     */
    public void setWeight( int s, int d, double w ) {
        if ( weights != null && weights.length > s && weights.length > d )
            weights[ s ][ d ] = w;        
    }
    
    /**
     * Get the weight value of a connection.
     * <P>
     * @param s source neuron
     * @param d destination neuron
     * @return weight, lowest double value, if there is no connection
     */
    public double getWeight( int s, int d ) {
        if ( weights != null && weights.length > s && weights.length > d )
            return weights[ s ][ d ];
        else
            return Double.longBitsToDouble( 0x1L );
    }
    
    /**
     * Initialize the weight matrix with values between -1.0 and 1.0. 
     * If there is a connection matrix, only the valid connections are set.
     * <P>
     * @param conMatrix connection matrix, maybe null
     */
    public void createRandomWeights( ConnectionMatrix conMatrix ) {
        if ( weights != null ) {
            for ( int i = 0; i < weights.length; i++ ) {
                for ( int j = 0; j < weights.length; j++ ) {
                    if ( conMatrix != null ) {
                        if ( conMatrix.isConnection( i, j ) )
                            weights[ i ][ j ] = Math.random() * 2.0 - 1.0;
                    } else 
                        weights[ i ][ j ] = Math.random() * 2.0 - 1.0;
                }
            }
        }
    }
     
    /**
     * Get the size of the matrix -1 for bias connections.
     * <P>
     * @return matrix size
     */
    public int size() {
        int retVal = 0;
        if ( weights != null )
            retVal = weights.length - 1;
        return retVal;
    }

    /**
     * Get a String representation of the matrix.
     * <P>
     * @return String representation
     */
    public String toString() {
        StringBuffer retVal = new StringBuffer( "Weight Matrix:\n\n" );
        for ( int i = 0; i < weights.length; i++ ) {
            if ( i == 0 )
                retVal.append( "B(" + i + ")\t" );
            else if ( i < input + 1 )
                retVal.append( "I(" + i + ")\t" );
            else if ( i < input + hidden + 1 )
                retVal.append( "H(" + i + ")\t" );
            else
                retVal.append( "O(" + i + ")\t" );
            for ( int j = 0; j < weights.length; j++ )
                retVal.append( weights[ i ][ j ] + "  " );
            retVal.append( "\n" );
        }
        return retVal.toString();
    }
}
