package edu.cosy.netgen.ga;

import java.beans.*;
import banda.genalg.BitChromosome;

import edu.cosy.netgen.ann.*;

/**
 * Genotype for encoding a ANN bean with four chromosome.
 */
public class ANNGenotypeExtended extends ANNGenotype {

    /**
     * Copy constructor for this genotype.
     * <P>
     * @param genotype genotype to copy
     */
    public ANNGenotypeExtended( ANNGenotypeExtended genotype ) {
        super( genotype );
    }

    /**
     * Create a new genotype with four chromosome. Connection matrix, 
     * training paramters and maximum training cycles, activation functions and neuron markers,
     * the weight matrix, each are encoded in one chromosome.     
     * <P>
     * @param bean master ANN bean for encoding
     * @param co connection matrix encoding?
     * @param tp training paramteres encoding?
     * @param cy maximum training cycles encoding?
     * @param ac activation function encoding?
     * @param ma neuron markers encoding?
     * @param wt weight matrix encoding?
     * @param minW minimum value for weights
     * @param maxW maximum value for weights
     * @param afIn pool of activation functions for input neurons
     * @param afHidden pool of activation functions for hidden neurons
     * @param afOut pool of activation functions for output neurons
     */
    public ANNGenotypeExtended( ANN bean, boolean co, boolean tp, boolean cy, boolean ac,
                                boolean ma, boolean wt, double minW, double maxW, 
                                String[] afIn, String[] afHidden, String[] afOut ) {    
        super( 4 );       
        this.bean = bean;
        actFunctionPoolIn = afIn;
        actFunctionPoolHidden = afHidden;
        actFunctionPoolOut = afOut;
        int length = 0;
        if ( co ) {
            isCO = true;
            length += getConMatrixBitLength( bean );
        }
        setChromosome( 0, new BitChromosome( length ) );
        length = 0;
        if ( tp ) {
            isTP = true;
            length += getTrainingParamBitLength( bean );
        }
        if ( cy ) {
            isCY = true;
            length += getMaxTrainingCyclesBitLength( bean );
        }
        setChromosome( 1, new BitChromosome( length ) );
        length = 0;
        if ( ac ) {
            isAC = true;
            length += getActFunctionBitLength( bean );
        }
        if ( ma ) {
            isMA = true;
            length += getMarkersBitLength( bean );
        }
        setChromosome( 2, new BitChromosome( length ) );
        length = 0;
        if ( wt ) {
            isWT = true;
            this.minW = minW;
            this.maxW = maxW;
            length += getWeightMatrixBitLength( bean, minW, maxW );
        }
        setChromosome( 3, new BitChromosome( length ) );
   }
    
    /**
     * Get the phenotype, an ANN bean, of this genotype.
     * <P>
     * @return decoded ANN bean, null if the genotype is emty
     */
    public Object getPhenotype() {
        ANN retVal = null;
        BitChromosome[] chroms = ( BitChromosome[] ) getChromosomes();
        if ( bean != null && ( chroms[ 0 ].length() > 0 || chroms[ 1 ].length() > 0 
                            || chroms[ 2 ].length() > 0 || chroms[ 3 ].length() > 0 ) ) {
            try {
                retVal = ( ANN ) Beans.instantiate( null, "edu.cosy.netgen.ann.ANN" );
                retVal.setDescription( bean.getDescription() );
                retVal.setTrainingFunction( bean.getTrainingFunction() );
                retVal.setNetType( bean.getNetType() );
                retVal.setIn( bean.getIn() );
                retVal.setHidden( bean.getHidden() );
                retVal.setOut( bean.getOut() );
                retVal.setStopError( bean.getStopError() );
                
                int ci = 0;

                // set connection matrix
                if ( isCO && chroms[ 0 ].length() > 0 )
                    retVal.setConMatrix( decodeConMatrix( chroms[ 0 ], ci ) ); 
                else
                    retVal.setConMatrix( new ConnectionMatrix( bean.getConMatrix() ) );    
                ci = 0;     

                // set training parameters
                if ( isTP && chroms[ 1 ].length() > 0 )
                    retVal.setTrainingFunctionParams( decodeTrainingParams( chroms[ 1 ], ci ) );
                else {
                    if ( bean.getTrainingFunctionParams() != null ) {
                        TrainingParam[] tmp = new TrainingParam[ bean.getTrainingFunctionParams().length ];
                        System.arraycopy( bean.getTrainingFunctionParams(), 0, tmp, 0, tmp.length );
                        retVal.setTrainingFunctionParams( tmp );
                    }
                }

                // set training cycles
                if ( isCY && chroms[ 1 ].length() > 0 ) {
                    retVal.setStopError( 0.0 );
                    retVal.setMaxTrainingCycles( decodeMaxTrainingCycles( chroms[ 1 ], ci ) );
                }
                else
                    retVal.setMaxTrainingCycles( bean.getMaxTrainingCycles() );
                ci = 0;

                // set activation functions    
                if ( isAC && chroms[ 2 ].length() > 0 )
                    retVal.setActivationFunctionsByNames( decodeActFunctions( chroms[ 2 ], ci ) );
                else {
                    if ( bean.getActivationFunctions() != null ) {
                        ActFunction[] tmp = new ActFunction[ bean.getActivationFunctions().length ];
                        System.arraycopy( bean.getActivationFunctions(), 0, tmp, 0, tmp.length );
                        retVal.setActivationFunctions( tmp );
                    }
                }
                
                // set neuron markers
                if ( isMA && chroms[ 2 ].length() > 0 )
                    retVal.setNeuronMarkers( decodeMarkers( chroms[ 2 ], ci ) );
                else {
                    if ( bean.getNeuronMarkers() != null ) {
                        boolean[] tmp = new boolean[ bean.getNeuronMarkers().length ];
                        System.arraycopy( bean.getNeuronMarkers(), 0, tmp, 0, tmp.length );
                        retVal.setNeuronMarkers( tmp );
                    }
                }                
                ci = 0;

                // set weight matrix
                if ( isWT && chroms[ 3 ].length() > 0 )
                    retVal.setWeightMatrix( decodeWeightMatrix( chroms[ 3 ], ci ) );
                else
                    retVal.setWeightMatrix( new WeightMatrix( bean.getWeightMatrix() ) );                

                retVal.setFitnessFunction( bean.getFitnessFunction() );
                retVal.setRegFunction( bean.getRegFunction() );
            } catch ( Exception e ) {
                return null;
            }
        }
        return retVal;
    }

    /**
     * Function for cloning.
     * <P>
     * @return new clone object
     */
    public Object clone() {
        return new ANNGenotypeExtended( this );
    }
}
