package edu.cosy.netgen.parser;
 
import java.io.*;
import java.text.*;
import java.util.*;
import java.beans.*;

import edu.cosy.netgen.ann.*;

public class AnnSNNSParser implements AnnParser {
    private final static String DESCRIPTION = "SNNS 4.2 Format Parser for netGen 0.3";
    private final static String[] NET_EXT = { "net", "snns" };
    private final static String[] PAT_EXT = { "pat" };
    
    public String getDescription() {
        return DESCRIPTION;
    }
    
    public String[] getNetFileExt() {
        return NET_EXT;
    }
    
    public String[] getPatternFileExt() {
        return PAT_EXT;
    }

    public PatternSet importPatterns( File file ) throws AnnParserException {
        PatternSet retVal = null;
        if ( file != null ) {
            int noPat = 0;
            int noInput = 0;
            int noOutput = 0;
            String str = new String();
            boolean flag = false;
            StringBuffer buffer = new StringBuffer();
            try {
                BufferedReader in = new BufferedReader( new FileReader( file ) );
                while ( ( str = in.readLine() ) != null ) {
                    if ( str.length() == 0 )
                        continue;
                    if ( str.charAt( 0 ) == '#' )
                        continue;
                    if ( str.startsWith( "No. of patterns :" ) ) {
                        noPat = Integer.parseInt( ( str.substring( str.indexOf( ':' ) + 1 ) ).trim() );
                        continue;
                    }
                    if ( str.startsWith( "No. of input units :" ) ) {
                        noInput = Integer.parseInt( ( str.substring( str.indexOf( ':' ) + 1 ) ).trim() );
                        flag = true;
                        continue;
                    }
                    if ( str.startsWith( "No. of output units :" ) ) {
                        noOutput = Integer.parseInt( ( str.substring( str.indexOf( ':' ) + 1 ) ).trim() );
                        continue;
                    }
    
          // starts reading numbers after praefix2 was found
                    if ( flag )
                        buffer.append( str );
                }
            }
            catch( NumberFormatException nfe ) {
                throw new AnnParserException( "Cannot read number in file pattern " + file + "!" );
            }
            catch( FileNotFoundException fnfe ) {
                throw new AnnParserException( "Pattern file " + file + " not found!" );
            }
            catch( Exception e ) {
                throw new AnnParserException( "Cannot read pattern file " + file + "!" );
            }
      
            StringTokenizer tokenizer = new StringTokenizer( buffer.toString() );
            Pattern psetin[] = new Pattern[ noPat ];
            Pattern psetout[] = new Pattern[ noPat ];
            int za = 0;
            for ( int j = 0; j < ( noPat * 2 ); j = j + 2 ) {
                
                // input units
                Pattern pat = new Pattern( noInput );
                for ( int i = 0; i < noInput; i++ )
                    pat.setValue(i, Double.parseDouble( tokenizer.nextToken() ) );
    
                // output units
                Pattern pat1 = new Pattern( noOutput );
                for ( int i = 0; i < noOutput; i++ )
                    pat1.setValue( i, Double.parseDouble( tokenizer.nextToken() ) );
                psetin[ za ] = pat;
                psetout[ za ] = pat1;
                za = za + 1;
            }

            retVal = new PatternSet( psetin, psetout );
        }
        return retVal;
    }

    public boolean exportPatterns( File file, PatternSet patterns ) throws AnnParserException {
        boolean retVal = false;
        if ( patterns != null && file != null ) {
            StringBuffer buffer = new StringBuffer( "SNNS pattern definition file V3.2" );
            try {

                //things to make a nice format of numbers
                SimpleDateFormat formatter = new SimpleDateFormat( "EEE MMM d H:mm:ss yyyy", Locale.US );
                DecimalFormatSymbols dfs = new DecimalFormatSymbols();
                dfs.setDecimalSeparator( '.' );
                DecimalFormat df = new DecimalFormat( "0.00000", dfs );

                //write header
                buffer.append( "\ngenerated at " ).append( formatter.format( new Date() ) );
                buffer.append( "\n\nNo. of patterns : " ).append( patterns.getUsedPatternNr() );
                buffer.append( "\nNo. of input units : " ).append( patterns.getInPatterns()[ 0 ].getSize() );
                buffer.append( "\nNo. of output units : " ).append( patterns.getOutPatterns()[ 0 ].getSize() );
                buffer.append( "\n" );

                for( int i = 0; i < patterns.getSize(); i++) {
                    if( !patterns.isMarker( i ) ) {
                        buffer.append( "\n" );
                        for( int j = 0; j < patterns.getInPatterns()[ 0 ].getSize(); j++ ) {
                            buffer.append( patterns.getInPatterns()[ i ].getValue(j) ).append( " " );
                        }
                        buffer.append( "\n" );
                        for( int k = 0; k < patterns.getOutPatterns()[ 0 ].getSize(); k++ ) {
                            buffer.append( patterns.getOutPatterns()[ i ].getValue( k ) ).append( " " );
                        }
                    }
                }

                //create file writer
                PrintWriter writer = new PrintWriter( new FileWriter( file ) );
                      writer.println( buffer.toString() );
                writer.close();
                retVal = true;
            }
            catch( FileNotFoundException fnfe ) {
                throw new AnnParserException( "Net file " + file + " not found!" );
            } catch( Exception e ) {
                throw new AnnParserException( "Cannot write net file " + file + "!" );
            }
        }
        return retVal;
    }

    //connections of a net are counted
    private int countConnections( boolean conMatrix[][] ) {
        int val = 0;
        for ( int i = 1; i < conMatrix.length; i++ ) {
            for ( int j = 1; j < conMatrix.length; j++ ) {
                if ( conMatrix[ i ][ j ] )
                    val++;
            }
        }
        return val;
    }

    //convertion between SNNS and netGen activation function names (netGen->SNNS)
    private String actConvert( ActFunction[] actFuncNeuron, int whichNeuron ) {
        
        //if the act. func. is unknown it is replaced by the identity function, user gets a warning
        String af = new String( "Act_Identity" );        
        if ( actFuncNeuron != null ) {
            if ( ActFunctionSigmoid.class.isInstance( actFuncNeuron[ whichNeuron ] ) )
                af = "Act_Logistic";
            else if ( ActFunctionTanH.class.isInstance( actFuncNeuron[ whichNeuron ] ) )
                af = "Act_TanH";           
        }
        return af;
    }

    //convertion between SNNS and netGen activation function names (SNNS->netGen)
    private ActFunction actConvert( String afName ) {
        
        //if the act. func. is unknown it is replaced by the identity function, user gets a warning
        ActFunction af = new ActFunctionLinear();
        if ( afName != null ) {
            if ( afName.equals( "Act_Logistic" ) )
                af = new ActFunctionSigmoid();
            else if ( afName.equals( "Act_TanH" ) )
                af = new ActFunctionTanH();
        }
        return af;
    }

    //proove for correct act. func. (import)
    private boolean correctActFunction( String actFunc ) {
        boolean correct = false;
        if ( actFunc.equals( "Act_Logistic" ) || actFunc.equals( "Act_TanH" ) || actFunc.equals( "Act_Identity" ) )
            correct = true;
        return correct;
    }

    //proove for correct out. func. (import)
    private boolean correctOutFunction( String outFunc ) {
        boolean correct = false;
        if ( outFunc.equals( "Out_Identity" ) )
            correct = true;
        return correct;
    }

    /**
     * Imports a feedforward SNNS-netfile and creates a bean
     * @param file SNNS file
     * @param markersModus bean with marker modus
     * @param ffnb feedforward net bean
     */
    public ANN importNet( File file ) throws AnnParserException {
        ANN ffnb = null;
        if ( file != null ) {           
            int in = 0;
            int hidden = 0;
            int out = 0;
            int iho = 0;
    
            ActFunction defaultActFunction = new ActFunctionSigmoid();
            String defaultOutFunction = new String( "" );
            boolean[][] conMatrix = new boolean[ 1 ][ 1 ];
            double[][] weightMatrix = new double[ 1 ][ 1 ];
            ActFunction actFuncNeuron[] = new ActFunction[ 1 ];
            
            //gets false when an unknown function is dedected
            boolean correctAct = true;
            boolean correctOut = true;
    
            try {
                
                // create file reader
                BufferedReader br = new BufferedReader( new FileReader( file ) );
    
                //one line of the netfile
                String tmp = new String();
                
                //shows the section the parser works on
                boolean isReadableUnitDef = false;
                boolean isReadableDefault = false;
                boolean isReadableCon = false;
                
                //there are always four lines from the section keyword to the first data
                int posUnitDef = 4;
                int posDefault = 4;
                int posConnection = 4;
                
                // vars that help navigate in the sections
                int posDef = 0;
                int posInDefault = 0;
                int posWeight = 1;
                
                //shows the actual pos. in the sct. func. vector
                int posActFunc = 1;
                
                //shows the actual column in the weight and connection matrix
                int column = 0;
    
                // do it for all lines
                while ( ( tmp = br.readLine() ) != null ) {
    
                    //number of all units gets parsed
                    if ( tmp.startsWith( "no. of units" ) ) {
                        iho = new Integer( tmp.substring( tmp.indexOf( ":" ) + 2, tmp.length() ) ).intValue();
                        
                        //create data which needs the number of units
                        conMatrix = new boolean[ 1 + iho ][ 1 + iho ];
                        weightMatrix = new double[ 1 + iho ][ 1 + iho ];
                        
                        actFuncNeuron = new ActFunction[ 1 + iho ];
                    }
    
                    //default values for units get parsed
                    if ( tmp.startsWith( "unit default" ) )
                        isReadableDefault = true;
                    if ( isReadableDefault ) {
                        
                        //four lines under the keyword
                        if ( posDefault == 0 ) {
                            
                            //line gets tokened
                            StringTokenizer strTok = new StringTokenizer( tmp );
                            String str;
                            while  ( strTok.hasMoreTokens() ) {
                                
                                // symbols like | show the beginning an the end of a new datafield
                                if ( strTok.nextToken().equals( "|" ) )
                                    posInDefault++;
                                    
                                //datafield 5: default act. function
                                if ( posInDefault == 5 ) {
                                    str = strTok.nextToken();
                                    defaultActFunction = actConvert( str );
                                    
                                    //proove whether it is a known act. func.
                                    correctAct = correctActFunction( str );
                                }
                                
                                //datafield 6: out function
                                if ( posInDefault == 6 ) {
                                    
                                    //netGen don't support out functions, so only an identity func. is accepted
                                    correctOut = correctOutFunction( strTok.nextToken() );
                                    
                                    //we are at the end of this section
                                    isReadableDefault = false;
                                }
    
                            }
                        }
                        if ( posDefault > 0 )
                            posDefault--;
                    }
    
                    // now we are in the unit definition section
                    if ( tmp.startsWith( "unit definition" ) )
                        isReadableUnitDef = true;
                    if ( isReadableUnitDef ) {
                        if ( posUnitDef == 0 ) {
    
                            StringTokenizer strTok = new StringTokenizer( tmp );
                            String str;
                            
                            //shows the actual datafield
                            posDef = 0;
                            while ( strTok.hasMoreTokens() ) {
                                str = strTok.nextToken();
                                
                                //we encountered a new datafield
                                if ( str.equals( "|" ) )
                                    posDef++;
                                    
                                //too lazy to write an extra out function, we are at the end of the line and so we break free
                                if ( str.equals( "||" ) ) {
                                    break;
                                }
                                
                                //neither an act. func. nor an out func. is written down, so default functions get used, and we are again at the and of the line and break
                                if ( str.equals( "|||" ) ) {
                                    actFuncNeuron[ posActFunc ] = defaultActFunction;
                                    posActFunc++;
                                    break;
                                }
    
                                //here the bias is written in the weight matrix
                                if ( posDef == 4 ) {
                                    weightMatrix[ 0 ][ posWeight ] = Double.parseDouble( strTok.nextToken() );
                                    posWeight++;
                                }
    
                                //number of in, hidden and out units are counted
                                if ( posDef == 5 ) {
                                    String strHelp;
                                    strHelp = strTok.nextToken();
                                    if ( strHelp.equals( "i" ) )
                                        in++;
                                    if ( strHelp.equals( "h" ) )
                                        hidden++;
                                    if ( strHelp.equals( "o" ) )
                                        out++;
                                }
                                
                                //the act. func. are written in the vector
                                if ( posDef == 7 ) {
                                    String strHelp;
                                    strHelp = strTok.nextToken();
                                    
                                    //no act. func. stands in the datafield
                                    if ( strHelp.equals( "|" ) ) {
                                        actFuncNeuron[ posActFunc ] = defaultActFunction;
                                        posActFunc++;
                                    }
                                    
                                    //an act. func exist for the actual neuron
                                    else {
                                        actFuncNeuron[ posActFunc ] = actConvert( strHelp );
                                        posActFunc++;
                                        correctAct = correctActFunction( strHelp );
                                    }
                                }
                                
                                //is the out func. correct?
                                if ( posDef == 8 ) {
                                    correctOut = correctOutFunction( strTok.nextToken() );
                                }
                                
                                // thank god, -- determines the end of the section
                                if ( tmp.startsWith( "--" ) )
                                    isReadableUnitDef = false;
                            }
                        }
                        if ( posUnitDef > 0 )
                            posUnitDef -= 1;
                    }
    
                    // here we have the connection section
                    if ( tmp.startsWith( "connection definition" ) )
                        isReadableCon = true;
                    if ( isReadableCon ) {
                        if ( posConnection == 0 ) {
                            
                            //we need a different StringTokenizer, : must also be a delimiter
                            StringTokenizer strTok = new StringTokenizer( tmp, "\t\n\r\f :" );
                            
                            //end of the section
                            if ( tmp.startsWith( "--" ) )
                                isReadableCon = false;
                            if ( isReadableCon ) {
                                String str;
                                String strHelp;
                                int row;
                                double weight;
    
                                //the first weight gets parsed, its outside the while because there must be a check whether there are only more weights in a new line or there is a new receiving neuron
                                str = strTok.nextToken();
                                strHelp = strTok.nextToken(); //System.out.println("DDD"+strHelp);
                                
                                //remove the ,
                                if ( strHelp.endsWith( "," ) )
                                    strHelp = strHelp.substring( 0, strHelp.length() - 1 );
                                    
                                //a new neuron which receives connections is encountered
                                if ( strHelp.equals( "|" ) ) {
                                    column = Integer.parseInt( str );
                                    str = strTok.nextToken(); //str=strTok.nextToken();
                                    
                                    //if a site is written down, sites are ignored by netGen
                                    if ( !( str.equals( "|" ) ) )
                                        str = strTok.nextToken();
                                }
                                
                                //no new neuron, only a new line with new connections
                                else {
                                    weightMatrix[ Integer.parseInt( str ) ][ column ] = Double.parseDouble( strHelp );
                                    conMatrix[ Integer.parseInt( str ) ][ column ] = true;
                                }
                                
                                //all connections, weights are parsed
                                while ( strTok.hasMoreTokens() ) {
                                    
                                    //look for the input neuron
                                    row = Integer.parseInt( strTok.nextToken() );
                                    
                                    //the weight
                                    str = strTok.nextToken();
                                    if ( str.endsWith( "," ) )
                                        str = str.substring( 0, str.length() - 1 );
                                    weightMatrix[ row ][ column ] = Double.parseDouble( str );
                                    conMatrix[ row ][ column ] = true;
                                }
                            }
                        }
                        if ( posConnection > 0 )
                            posConnection -= 1;
                    }
                } //for
    
                // close file reader
                br.close();
    
                ffnb = ( ANN ) Beans.instantiate( null, "edu.cosy.netgen.ann.ANN" );
                ffnb.setNetType( NeuralNet.FEEDFORWARD );
                ffnb.setIn( in );
                ffnb.setHidden( hidden );
                ffnb.setOut( out );
                ConnectionMatrix tmpConMatrix = new ConnectionMatrix( in, hidden, out, ConnectionMatrix.CREATE_BIAS_ONLY );
                for ( int i = 0; i < conMatrix.length; i++ )
                    for ( int j = 0; j < conMatrix[ 0 ].length; j++ )
                        if ( conMatrix[ i ][ j ] )
                            tmpConMatrix.setConnection( i, j );
                ffnb.setConMatrix( tmpConMatrix );
                WeightMatrix tmpWeightMatrix = new WeightMatrix( in, hidden, out );
                for ( int i = 0; i < weightMatrix.length; i++ )
                    for ( int j = 0; j < weightMatrix[ 0 ].length; j++ )
                        if ( conMatrix[ i ][ j ] )
                            tmpWeightMatrix.setWeight( i, j, weightMatrix[ i ][ j ] );
                ffnb.setWeightMatrix( tmpWeightMatrix );
                ffnb.setActivationFunctions( actFuncNeuron );
            }
            catch( NumberFormatException nfe ) {
                throw new AnnParserException( "Cannot read number in net file " + file + "!" );
            }
            catch( FileNotFoundException fnfe ) {
                throw new AnnParserException( "Net file " + file + " not found!" );
            }
            catch( Exception e ) {
                throw new AnnParserException( "Cannot read net file " + file + "!" );
            }
        }
        return ffnb;
    }

    /**
     * Exports a feedforward SNNS-netfile
     * @param file output file
     * @param ffnb feedforward net bean
     */
    public boolean exportNet( File file, ANN ffnb )  throws AnnParserException {
        boolean retVal = false;
        if ( ffnb != null && file != null )
        {
            StringBuffer buffer = new StringBuffer( "SNNS network definition file V1.4-3D" );
            String netName = new String( "netGen 0.3 Export-Net" );
    
            String learningFunction = new String( ffnb.getTrainingFunctionByName() );
            String updateFunction = new String( "Topological_Order" );
    
            int in = ffnb.getIn();
            int hidden = ffnb.getHidden();
            int out = ffnb.getOut();
            ConnectionMatrix conMatrix = ffnb.getConMatrix();
            WeightMatrix weightMatrix = ffnb.getWeightMatrix();
            ActFunction[] actFuncNeuron = ffnb.getActivationFunctions();
        
            try {
    
                //things to make a nice format of numbers
                SimpleDateFormat formatter = new SimpleDateFormat( "EEE MMM d H:mm:ss yyyy", Locale.US );
                DecimalFormatSymbols dfs = new DecimalFormatSymbols();
                dfs.setDecimalSeparator( '.' );
                DecimalFormat df = new DecimalFormat( "0.00000", dfs );
    
                //write header
                buffer.append( "\ngenerated at " ).append( formatter.format( new Date() ) );
                buffer.append( "\n\nnetwork name : " ).append( netName );
                buffer.append( "\nsource files :\nno. of units : " ).append( in + hidden + out );
                buffer.append( "\nno. of connections : " ).append( conMatrix.countConnections() );
                buffer.append( "\nno. of unit types : 0\nno. of site types : 0" );
    
                int no = 0;
                double act = 0.00000;
                double bias = 0.00000;
                int za1 = 0;
    
                //write default section
                buffer.append( "\n\nlearning function : " ).append( learningFunction );
                buffer.append( "\nupdate function   : " ).append( updateFunction );
                buffer.append( "\n\nunit default section :" );
                buffer.append( "\n\nact      | bias     | st | subnet | layer | act func     | out func" );
                buffer.append( "\n---------|----------|----|--------|-------|--------------|-------------" );
                buffer.append( "\n 0.00000 |  0.00000 | h  |      0 |     1 | Act_Logistic | Out_Identity " );
                buffer.append( "\n---------|----------|----|--------|-------|--------------|-------------" );
                buffer.append( "\n\nunit definition section :" );
                buffer.append( "\n\nno. | typeName | unitName | act      | bias     | st | position | act func | out func | sites" );
                buffer.append( "\n----|----------|----------|----------|----------|----|----------|----------|----------|-------" );
    
                //writes the used act. func. of the input neurons, in original SNNS-files df.format(act) would show one pattern+output, bias is always 0 for input neurons
                for ( int za = 1; za < 1 + in; za++ ) {
                    no++;
                    za1++;
                    buffer.append( "\n" ).append( no ).append( "\t|          |          | " )
                          .append( df.format( act ) ).append( "  | " ).append( df.format( bias ) )
                          .append( "\t| i  |  0, " ).append( za1 ).append( ", 0 | " )
                          .append( actConvert( actFuncNeuron, za ) ).append( " ||" );

                }
    
                //writes all lines with hidden units, bias, act. func.
                for ( int za = 1 + in; za < 1 + in + hidden; za++ ) {
                    no++;
                    za1++;
                    buffer.append( "\n" ).append( no ).append( "\t|          |          | " )
                          .append( df.format( act ) ).append( "  | " ).append( df.format( weightMatrix.getWeight( 0, za ) ) )
                          .append( "\t| h  |  0, " ).append( za1 ).append( ", 0 | " )
                          .append( actConvert( actFuncNeuron, za ) ).append( " ||" );
                }
    
                //writes all lines with output units, bias, act.func.
                for ( int za = 1 + in + hidden; za < 1 + in + hidden + out; za++ ) {
                    no++;
                    za1++;
                    buffer.append( "\n" ).append( no ).append( "\t|          |          | " )
                          .append( df.format( act ) ).append( "  | " ).append( df.format( weightMatrix.getWeight( 0, za ) ) )
                          .append( "\t| o  |  0, " ).append( za1 ).append( ", 0 | " )
                          .append( actConvert( actFuncNeuron, za ) ).append( " ||" );
                }
                buffer.append( "\n----|----------|----------|----------|----------|----|----------|----------|----------|-------" );
                
                //writes the connections
                buffer.append( "\n\nconnection definition section :" );
                buffer.append( "\n\ntarget | site | source:weight" );
                buffer.append( "\n-------|------|---------------------------------------------------------------------------------------------------------------------" );
    
                //writes the target units, sources and weights
                for ( int i = 1 + in; i < 1 + in + hidden + out; i++ ) {
                    
                    if ( conMatrix.isAnyConnection( i ) ) {
                        StringBuffer sb = new StringBuffer( "\n\t" );
                        sb.append( i ).append( " |      |  " );
                        int zah = 0;
                        for ( int j = 1; j < 1 + in + hidden + out; j++ ) {
                            
                            //only sources which are alive are written down
                            if ( conMatrix.isConnection( j, i ) ) {
                                zah++;
                                
                                //only 9 weights are written in one line
                                if ( zah == 10 ) {
                                    buffer.append( sb );
                                    sb = new StringBuffer( "                " );
                                    zah = 1;
                                }
                                sb.append( j ).append( ": " ).append( df.format( weightMatrix.getWeight( j, i ) ) ).append( ", " );
                            }
                        }
                        
                        //necessary to eliminate the last ", "
                        int sbLength = sb.length();
                        if ( sb.substring( sbLength - 2, sbLength ).equals( ", " ) ) {
                            sb.setLength( sbLength - 2 );
                            buffer.append( sb );
                        }
                    }
                }
                buffer.append( "\n-------|------|---------------------------------------------------------------------------------------------------------------------" );
    
                //create file writer
                PrintWriter writer = new PrintWriter( new FileWriter( file ) );
                writer.println( buffer.toString() );
                writer.close();
                retVal = true;
            }
            catch( FileNotFoundException fnfe ) {
                throw new AnnParserException( "Net file " + file + " not found!" );
            }
            catch( Exception e ) {
                throw new AnnParserException( "Cannot write net file " + file + "!" );
                }
        }
        return retVal;
    }
}
