package edu.cosy.netgen.server;

import java.util.Vector;

import banda.genalg.Population;
import edu.cosy.netgen.ga.Individual;
import edu.cosy.netgen.server.event.ContainerEvent;
import edu.cosy.netgen.server.event.ContainerListener;

/**
 * Container that stores the genotypes for distribution.
 */
public class Container {
    private Population currentPopulation;
    
    // listeners for container events
    private Vector listeners;
    
    // type of current task
    private int taskType = -1;
    
    // counter for calculated genotypes
    private int readyCounter = 0;
    
    // core container
    private Individual container[];
    
    /**
     * Initialize the container with a specific population.
     * <P>
     * @param pop population for distribution
     */
    public void initialize(  Population pop ) {
        if ( pop != null ) {
            currentPopulation = pop;           
            readyCounter = 0;
            container = new Individual[ pop.size() ];
            for ( int i = 0; i < pop.size(); i++ )
                container[ i ] = new Individual( i, pop.genotypeAt( i ) );
            notifyInitializedEvent( new ContainerEvent( this ) );
        }
    }
    
    /**
     * Set the fitness value of a genotype in the container.
     * <P>
     * @param i index of the genotype in container
     * @param v fitness value
     */
    public synchronized void setFitness( int i, double v ) {
        if ( container != null ) {
            container[ i ].setFitness( v );
            container[ i ].isCalculated = true;
            container[ i ].isDistributed = false;
            readyCounter++;
            if ( readyCounter == container.length ) 
                notifyCalculatedEvent( new ContainerEvent( this ) );
        }
    }             

    /**
     * Get the next valid genotype for calculation.
     * <P>
     * @return individual to be calculated
     */
    public synchronized Individual getNextIndividual() {
        Individual retVal = null;
        for ( int i = 0; i < container.length; i++ ) {
            if ( !container[ i ].isCalculated && !container[ i ].isDistributed ) {
                container[ i ].isDistributed = true;
                retVal = container[ i ];
                notifyIndividualDistributedEvent( new ContainerEvent( this, retVal ) );
                break;
            }
        }
        return retVal;
    }
    
    /**
     * Set the current task type.
     * <P>
     * @param type task type
     */
    public synchronized void setCurrentTaskType( int type ) {
        taskType = type;
    }

    /**
     * Get the current task type.
     * <P>
     * @return task type
     */
    public synchronized int getCurrentTaskType() {
        return taskType;
    }
    
    /**
     * Add a listener for the container.
     * <P>
     * @param l new listener
     */
    public synchronized void addListener( ContainerListener l ) {
        if ( listeners == null )
            listeners = new Vector();
        listeners.add( l );
    }
    
    /**
     * Remove a listener from the container.
     * <P>
     * @param l listener to be removed
     */
    public synchronized void removeListener( ContainerListener l ) {
        if ( listeners != null && l != null )
            listeners.remove( l );
    }
    
    /**
     * Get all listneres of the container.
     * <P>
     * @return array of listeners
     */
    public synchronized ContainerListener[] getListeners() {
        Vector list;
        ContainerListener[] retVal = null;
        if ( listeners != null ) {
            synchronized ( this ) {
                list = ( Vector ) listeners.clone();
            }
            retVal = ( ContainerListener[] ) list.toArray();
        }
        return retVal;
    }    
    
    /**
     * Notify all listeners that the container is initialized.
     * <P>
     * @param e event to be notified
     */
    public void notifyInitializedEvent( ContainerEvent e ) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                ContainerListener listener = ( ContainerListener ) list.elementAt( i );
                listener.containerInitialized( e );
            }
        }
    }
    
    /**
     * Notify all listeners that all genotypes are calculated.
     * <P>
     * @param e event to be notified
     */
    public void notifyCalculatedEvent( ContainerEvent e ) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                ContainerListener listener = ( ContainerListener ) list.elementAt( i );
                listener.containerCalculated( e );
            }
        }
    }
    
    /**
     * Notify all listeners that a genotype is distributed for calculation.
     * <P>
     * @param e event to be notified
     */
    public void notifyIndividualDistributedEvent( ContainerEvent e ) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                ContainerListener listener = ( ContainerListener ) list.elementAt( i );
                listener.containerIndividualDistributed( e );
            }
        }
    }
}
