package edu.cosy.netgen.server;

import java.io.*;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Vector;
import java.util.Properties;
import fr.dyade.koala.xml.koml.KOMLSerializer;
import fr.dyade.koala.xml.koml.KOMLDeserializer;

/**
 * Implementation of the repository that stores in files.
 * For XML, this parser was validated with IBM XML-Parser xml4J_3_0_0EA3.
 */
public class DiskRepository extends Repository {
        
    // directory for data storage
    private File storeDir;
            
    /**
     * Default constructor for the repository. The following properties are set:
     * <P>
     * repository.storeDir   = data/               ( directory there the data items are stored )
     * repository.taskExt    = task                ( extension of the task files )
     * repository.wcListFile = netgen.wclients     ( file there the working client information is stored )
     * repository.useXML     = true                ( flag for using XML at object serialization )
     */
    public DiskRepository() {
        super();
        properties.addProperty( propKey + ".storeDir", "data/" );
        properties.addProperty( propKey + ".taskExt", "task" );
        properties.addProperty( propKey + ".wcListFile", "netgen.wclients" );
        properties.addBooleanProperty( propKey + ".useXML", true );
    }
    
    /**
     * Initialize the file repository.
     * <P>
     * @return true, if the repository was successfully created
     */
    protected boolean initRepository() {
        storeDir = new File( properties.getProperty( propKey + ".storeDir" ) );
        if ( storeDir != null )
            return true;
        else
            return false;           
    }    
    
    /**
     * Checks if a specific task is stored in the repository.
     * <P>
     * @param id task ID to be checked
     * @return true, if the task is stored in the repository
     */
    public boolean isTaskStored( long id ) {
        return existItem( id, properties.getProperty( propKey + ".taskExt" ) );
    }
        
    /**
     * Get a list of IDs from the tasks in the repository.
     * <P>
     * @return array of stored task IDs
     */
    public long[] listStoredTasks() {
        return getStoredItemIDs( properties.getProperty( propKey + ".taskExt" ) );
    }
    
    /**
     * Store a task in the repository.
     * <P>
     * @param task task to be stored
     * @return true, if storing was successfully
     */
    public boolean storeTask( Task task ) {
        boolean retVal = false;
        if ( task != null )
            retVal = storeItem( task.getTaskID(), task, properties.getProperty( propKey + ".taskExt" ) );        
        return retVal;
    }
        
    /**
     * Load a task form the repository.
     * <P>
     * @param id ID of the task to be loaded
     * @return the task, null if a error occured
     */
    public Task loadTask( long id ) {
        Task retVal = null;
        if ( existItem( id, properties.getProperty( propKey + ".taskExt" ) ) )
            retVal = ( Task ) loadItem( id, properties.getProperty( propKey + ".taskExt" ) );
        return retVal;
    }
          
    /**
     * Delete a task from the repository.
     * <P>
     * @param id ID of the task to be removed
     * @return true, if success
     */
    public boolean deleteStoredTask( long id ) {
        return deleteItem( id, properties.getProperty( propKey + ".taskExt" ) );
    }
        
    /**
     * Store the list of working client information.
     * <P>
     * @param list working client informations
     * @return true, if storing was successful
     */
    public boolean storeWorkingClientList( WorkingClientInfo[] list ) {
        boolean retVal = false;
        String tmp = properties.getProperty( propKey + ".wcListFile" );
        if ( list != null && tmp != null ) {
            int i = tmp.indexOf( '.' );
            retVal = storeItem( tmp.substring( 0, i ), list, tmp.substring( i + 1, tmp.length() ) );
        }
        return retVal;
    }

    /**
     * Load the list with working client information.
     * <P>
     * @return array of working client info
     */
    public WorkingClientInfo[] loadWorkingClientList() {
        WorkingClientInfo[] retVal = null;
        String tmp = properties.getProperty( propKey + ".wcListFile" );
        if ( tmp != null && ( new File( tmp ) ).exists() ) {
            int i = tmp.indexOf( '.' );
            retVal = ( WorkingClientInfo[] ) loadItem( tmp.substring( 0, i ), tmp.substring( i + 1, tmp.length() ) );
        }
        return retVal;
    }    
        
    /**
     * Close the file repository.
     */
    public void closeRepository() {
    }

    // store item
    private boolean storeItem( long id, Serializable item, String ext ) {
        return storeItem( Long.toString( id ), item, ext );
    }

    private boolean storeItem( String fn, Serializable item, String ext ) {
        boolean retVal = false;
        if ( storeDir != null ) {            
            try {
                if ( !storeDir.exists() )
                    storeDir.mkdir();
                File file = new File( storeDir, fn + "." + ext );
                if ( !file.exists() )
                    file.createNewFile();
                FileOutputStream out = new FileOutputStream( file );
                if ( properties.getBooleanProperty( propKey + ".useXML" ) ) {
                    KOMLSerializer serializer = new KOMLSerializer( out, false );
                    serializer.addObject( item ); 
                    serializer.close();
                } else {
                    ObjectOutputStream s = new ObjectOutputStream( out );
                    if ( s != null ) {
                        s.writeObject( item );
                        s.flush();
                        s.close();
                    }
                }
                if ( out != null )
                    out.close();   
                retVal = true; 
            } catch ( Exception e ) {
                CAT.error( "Cannot store " + storeDir + "/" + fn + "." + ext + "! " + e );    
            }
        }
        return retVal;
    }
    
    // load item
    private Object loadItem( long id, String ext ) {
        return loadItem( Long.toString( id ), ext );
    }

    private Object loadItem( String fn, String ext ) {
        Object retVal = null;
        if ( storeDir != null && storeDir.exists() ) {
            try {
                File file = new File( storeDir, fn + "." + ext );
                if ( file.exists() ) {
                    FileInputStream in = new FileInputStream( file );
                    if ( properties.getBooleanProperty( propKey + ".useXML" ) ) {
                        KOMLDeserializer deserializer = new KOMLDeserializer( in, false );
                        try {
                            retVal = deserializer.readObject();
                        } catch ( EOFException eofe ) {
                            CAT.error( "ERROR: Cannot deserialize " + storeDir + "/" + fn + "." + ext + "! " + eofe );    
                        } finally {
                            deserializer.close();
                        }
                    } else {
                        ObjectInputStream s = new ObjectInputStream( in );
                        if ( s != null ) {
                            retVal = s.readObject();
                            s.close();
                        }
                    }
                    if ( in != null )
                        in.close();
                }
            } catch ( Exception e ) {
                CAT.error( "ERROR: Cannot read " + storeDir + "/" + fn + "." + ext + "! " + e );    
                retVal = null;
            }
        }
        return retVal;
    }

    // check existence
    private boolean existItem( long id, String ext ) {
        File item = new File( storeDir, id + "." + ext );
        return item.exists();
    }
        
    // list stored items
    private long[] getStoredItemIDs( String ext ) {
        long[] retVal = null;
        if ( storeDir != null && storeDir.exists() ) {
            String[] tmp = storeDir.list();
            retVal = new long[ tmp.length ];
            for ( int i = 0; i < tmp.length; i++ )
                if ( tmp[ i ].endsWith( "." + ext ) )
                    retVal[ i ] = Long.parseLong( tmp[ i ].substring( 0, tmp[ i ].indexOf( "." + ext ) ) );
            if ( retVal.length == 0 )
                retVal = null;
        }
        return retVal;
    }

    // delete item
    private boolean deleteItem( long id, String ext ) {
        boolean retVal = false;
        File file = new File( storeDir, id + "." + ext );
        if ( file.exists() )
            retVal = file.delete();
        return retVal;
    }
}
